from loguru import logger as log
from abc import ABC
from typing import Union, Type
from pydantic import BaseModel


def _read_normal_file_data(file_path: str, encoding: str = 'utf-8') -> str:
    with open(file_path, encoding=encoding) as f:
        config_data = f.read()
    return config_data


class DataFileLoader(ABC):
    """数据/文件加载器, 可将str, bytes类型数据或文件数据加载为python基本数据类型或pydantic模型"""

    def load(self,
             config_data: Union[str, bytes],
             encoding: str = 'utf-8',
             modelclass: Union[Type[dict], Type[BaseModel]] = dict,
             **kwargs) -> Union[dict, list, BaseModel]:
        """解析数据字符串到python dict或pydantic.BaseModel子类实例中

        Args:
            config_data: 数据字符串
            encoding: 数据编码类型, config_data 为 bytes 类型需传
            modelclass: 解析的每个记录的模型类型
        """
        model_data = self._convert_config_data_to_python(
            config_data, encoding=encoding, **kwargs)
        if isinstance(model_data, (tuple, list)):
            for i in range(len(model_data)):
                model_data[i] = self._convert_item_to_model(
                    model_data[i], modelclass=modelclass)
            return model_data
        return self._convert_item_to_model(model_data, modelclass=modelclass)

    def load_file(self,
                  file_path: str,
                  encoding: str = 'utf-8',
                  modelclass: Union[Type[dict], Type[BaseModel]] = dict,
                  **kwargs) -> Union[dict, list, BaseModel]:
        """解析文件数据字符串到python dict或pydantic.BaseModel子类实例中

        Args:
            file_path: 文件路径
            encoding: 文件编码
            modelclass: 解析的每个记录的模型类型
        """
        config_data = _read_normal_file_data(
            file_path=file_path, encoding=encoding)
        return self.load(config_data, encoding=encoding, modelclass=modelclass, **kwargs)

    def _convert_item_to_model(self, item: Union[dict, tuple, list],
                               modelclass: Union[Type[dict], Type[list], Type[tuple], Type[BaseModel]]):
        """将单个元素从dict, tuple, list 转换为指定的 modelclass

        Args:
            item: 单个元素类型
            modelclass: 解析的每个记录的模型类型
        """
        if isinstance(item, (tuple, list)):
            # 数组无法转直接返回
            if modelclass not in (tuple, list):
                log.warning(
                    '数据为tuple或list类型, 无法转为指定modelclass: %s, 直接返回', modelclass)
            return item
        if item and isinstance(item, dict):
            return modelclass(**item)
        return item

    def _convert_config_data_to_python(self,
                                       config_data: Union[str, bytes],
                                       encoding: str = 'utf-8',
                                       **kwargs) -> Union[dict, tuple, list]:
        """将数据字符串转为python对象

        Args:
            config_data: 数据
        """
        raise NotImplementedError()



from typing import Union

from .base import DataFileLoader


class CsvDataFileLoader(DataFileLoader):

    def _convert_config_data_to_python(self,
                                       config_data: Union[str, bytes],
                                       header=False,
                                       use_header_as_key=False,
                                       **kwargs) -> Union[dict, tuple, list]:
        """
        Args:
            header: 是否第一行为csv表格头
            use_header_as_key: 是否将header头作为key, 仅header = True生效
        """
        if isinstance(config_data, bytes):
            config_data = config_data.decode('utf-8')
        config_data = config_data.split('\n')
        keys = None
        if header:
            if len(config_data) <= 1:
                return []
            if use_header_as_key:
                keys = config_data[0].split(',')
            config_data = config_data[1:]
        if len(config_data) <= 0:
            return []
        result = []
        for config_data_item in config_data:
            if not config_data_item:
                continue
            config_data_item = config_data_item.split(',')
            if keys:
                item = {}
            else:
                item = []
            for i in range(len(config_data_item)):
                if keys:
                    item[keys[i]] = config_data_item[i]
                else:
                    item.append(config_data_item[i])
            result.append(item)
        return result



import asyncio

from dragons96_tools.cache import Cache, AsyncCache
from cachetools import TTLCache
from loguru import logger
from dragons96_tools.models import BaseModel


cache = Cache(cache=TTLCache(maxsize=100, ttl=60))
async_cache = AsyncCache(cache=TTLCache(maxsize=100, ttl=60))


class D(BaseModel):
    a: int


@cache.cacheable(key='func1_${a}_${b}')
def func1(a: int, b: str):
    return f'{a}_{b}'


@cache.cache_evict(key='func1_${a}_${b}')
def clear_func1_cache(a: int, b: str):
    pass


@cache.cache_put(key='func1_${a}_${b}')
def put_func1_cache(a: int, b: str):
    return f'{a}_{b}'


def test_cache():
    v1 = put_func1_cache(1, 's')
    v1 = func1(1, 's', D(a=123))
    logger.info("v1: {}", v1)
    clear_func1_cache(1, 's')
    v2 = func1(1, 's')
    logger.info("v2: {}", v2)


@async_cache.cacheable(key='func1_${a}_${b}')
async def async_func1(a: int, b: str):
    return f'{a}_{b}'


@async_cache.cache_evict(key='func1_${a}_${b}')
async def async_clear_func1_cache(a: int, b: str):
    pass


@async_cache.cache_put(key='func1_${a}_${b}')
async def async_put_func1_cache(a: int, b: str):
    return f'{a}_{b}'


def test_async_cache():
    v1 = asyncio.run(async_put_func1_cache(1, 's'))
    v1 = asyncio.run(async_func1(1, 's'))
    logger.info("v1: {}", v1)
    asyncio.run(async_clear_func1_cache(1, 's'))
    v2 = asyncio.run(async_func1(1, 's'))
    logger.info("v2: {}", v2)





from dragons96_tools.models import BaseModel
from typing import Optional, Type, List, Tuple, Any, Union
from pymysql import connect as c, Connection
from pymysql.cursors import DictCursor
from contextlib import contextmanager


def connect(host=None,
            port=3306,
            user=None,
            password='',
            database=None,
            cursorclass=DictCursor,
            autocommit=True,
            pool=False,
            pool_size=1,
            **kwargs):
    """业务Pymysql连接通用参数封装

    Returns:
        pymysql.Connection or dbutils.pooled_db.PooledDB
    """
    if pool:
        from dbutils.pooled_db import PooledDB
        return PooledDB(creator=connect,
                        mincached=pool_size,
                        host=host,
                        port=port,
                        user=user,
                        password=password,
                        database=database,
                        cursorclass=cursorclass,
                        autocommit=autocommit,
                        **kwargs)
    return c(host=host,
             port=port,
             user=user,
             password=password,
             database=database,
             cursorclass=cursorclass,
             autocommit=autocommit,
             **kwargs)


class _PyMysqlConfig(BaseModel):
    host: str
    port: int
    user: str
    password: Optional[str] = ''
    db: Optional[str] = None
    autocommit: Optional[bool] = False


class _PyMysqlTransactionCursor:
    """pymysql 事务 cursor 拓展, 仅执行 SQL 不提交不回滚, 由 PyMysqlHelper 处理提交回滚连接"""

    def __init__(self, conn: Connection):
        self._conn = conn
        self._cursor = conn.cursor()

    def execute(self,
                sql: str,
                args: Union[List[Any], Tuple[Any], None] = None,
                modelclass: Optional[Type[BaseModel]] = None
                ):
        """以事务方式执行 sql, 保障同一个 对象 执行多次 execute 方法在同一事务中

        Args:
            sql: 执行的sql
            args: 执行的sql中的参数
            modelclass: pydantic model 类型, 传入该值会将结果转为 pydantic model
        """
        self._cursor.execute(sql, args)
        ans = self._cursor.fetchall()
        if ans and modelclass:
            ans = list(map(lambda e: modelclass(**e), ans))
        return ans

    def close(self):
        if self._cursor:
            self._cursor.close()


class PyMysqlClient:
    """pymysql 客户端"""

    def __init__(self,
                 host='localhost',
                 port=3306,
                 user='root',
                 password='',
                 database=None,
                 **kwargs):
        self._config = _PyMysqlConfig(
            host=host,
            port=port,
            user=user,
            password=password,
            db=database
        )
        self._kw = kwargs

    def _new_connection(self):
        return c(host=self._config.host,
                 port=self._config.port,
                 user=self._config.user,
                 password=self._config.password,
                 db=self._config.db,
                 autocommit=self._config.autocommit,
                 **self._kw)

    def execute(self,
                sql: str,
                args: Union[List[Any], Tuple[Any], None] = None,
                modelclass: Optional[Type[BaseModel]] = None) -> Union[
        List[dict], List[BaseModel], Tuple[dict], Tuple[BaseModel]]:
        """执行 非事务模式 sql

        Args:
            sql: 执行的sql
            args: 执行的sql中的参数
            modelclass: pydantic model 类型, 传入该值会将结果转为 pydantic model
        """
        conn = self._new_connection()
        cursor = conn.cursor(cursor=DictCursor)
        try:
            cursor.execute(sql, args)
            conn.commit()
            ans = cursor.fetchall()
            if ans and modelclass:
                ans = list(map(lambda e: modelclass(**e), ans))
            return ans
        except Exception as e:
            conn.rollback()
            raise e
        finally:
            if cursor:
                cursor.close()
            if conn:
                conn.close()

    @contextmanager
    def transaction_client(self) -> _PyMysqlTransactionCursor:
        """获取事务游标对象

        使用方式
            helper = PyMysqlHelper(...)
            with helper.cursor() as cursor:
                users = cursor.execute('select * from user')
                cursor.execute('update user set user_name= "sss" where user_id=1')

        """
        conn = self._new_connection()
        transaction_cursor = _PyMysqlTransactionCursor(conn)
        try:
            yield transaction_cursor
            conn.commit()
        except Exception as e:
            conn.rollback()
            raise e
        finally:
            transaction_cursor.close()
            if conn:
                conn.close()
